import pytest
import pdb
from canoser import *
from datetime import datetime
import struct
import json

# Test Example in https://github.com/libra/libra/tree/master/common/canonical_serialization/README.md

ADDRESS_LENGTH = 32
ED25519_PUBLIC_KEY_LENGTH = 32
ED25519_SIGNATURE_LENGTH = 64

class Address(DelegateT):
    delegate_type = [Uint8, ADDRESS_LENGTH]


class ByteArray(DelegateT):
    delegate_type = [Uint8]


class TransactionArgument(RustEnum):
    _enums = [
        ('U64', Uint64),
        ('Address', Address),
        ('String', str),
        ('ByteArray', ByteArray)
    ]

class WriteOp(RustEnum):
    _enums = [
        ('Deletion', None),
        ('Value', ByteArray)
    ]

class AccessPath(Struct):
    _fields = [
        ('address', Address),
        ('path', ByteArray)
    ]


class Program(Struct):
    _fields = [
        ('code', [Uint8]),
        ('args', [TransactionArgument]),
        ('modules', [[Uint8]])
    ]


# `WriteSet` contains all access paths that one transaction modifies. Each of them is a `WriteOp`
# where `Value(val)` means that serialized representation should be updated to `val`, and
# `Deletion` means that we are going to delete this access path.
class WriteSet(Struct):
    _fields = [
        ('write_set', [(AccessPath, WriteOp)])
    ]


class Module(Struct):
    _fields = [
        ('code', [Uint8])
    ]


class Script(Struct):
    _fields = [
        ('code', [Uint8]),
        ('args', [TransactionArgument])
    ]

class TransactionPayload(RustEnum):
    _enums = [
        ('Program', Program),
        ('WriteSet', WriteSet),
        ('Script', Script),
        ('Module', Module)
    ]

class RawTransaction(Struct):
    _fields = [
        ('sender', Address),
        ('sequence_number', Uint64),
        ('payload', TransactionPayload),
        ('max_gas_amount', Uint64),
        ('gas_unit_price', Uint64),
        ('expiration_time', Uint64)
    ]


def test_readme_example1():
    lcs = '0200000020000000A71D76FAA2D2D5C3224EC3D41DEB293973564A791E55C6782BA76C2BF0495F9A2100000001217DA6C6B3E19F1825CFB2676DAECCE3BF3DE03CF26647C78DF00B371B25CC970000000020000000C4C63F80C74B11263E421EBF8486A4E398D0DBC09FA7D4F62CCDB309F3AEA81F0900000001217DA6C6B3E19F180100000004000000CAFED00D'
    tx = WriteSet.deserialize(bytes.fromhex(lcs))
    print(tx)

def test_readme_example2():
    lcs = '00000000040000006D6F766502000000020000000900000043414645204430304402000000090000006361666520643030640300000001000000CA02000000FED0010000000D'
    tx = TransactionPayload.deserialize(bytes.fromhex(lcs))
    print(tx)

def test_readme_example3():
    lcs = '010000000200000020000000A71D76FAA2D2D5C3224EC3D41DEB293973564A791E55C6782BA76C2BF0495F9A2100000001217DA6C6B3E19F1825CFB2676DAECCE3BF3DE03CF26647C78DF00B371B25CC970000000020000000C4C63F80C74B11263E421EBF8486A4E398D0DBC09FA7D4F62CCDB309F3AEA81F0900000001217DA6C6B3E19F180100000004000000CAFED00D'
    tx = TransactionPayload.deserialize(bytes.fromhex(lcs))
    print(tx)

def test_readme_example4():
    lcs = '200000003A24A61E05D129CACE9E0EFC8BC9E33831FEC9A9BE66F50FD352A2638A49B9EE200000000000000000000000040000006D6F766502000000020000000900000043414645204430304402000000090000006361666520643030640300000001000000CA02000000FED0010000000D1027000000000000204E0000000000008051010000000000'
    tx = RawTransaction.deserialize(bytes.fromhex(lcs))
    assert tx.__str__() == """{
  "sender": "3a24a61e05d129cace9e0efc8bc9e33831fec9a9be66f50fd352a2638a49b9ee",
  "sequence_number": 32,
  "payload": {
    "Program": {
      "code": "6d6f7665",
      "args": [
        {
          "String": "CAFE D00D"
        },
        {
          "String": "cafe d00d"
        }
      ],
      "modules": [
        "ca",
        "fed0",
        "0d"
      ]
    }
  },
  "max_gas_amount": 10000,
  "gas_unit_price": 20000,
  "expiration_time": 86400
}"""

def test_readme_example5():
    lcs = '20000000C3398A599A6F3B9F30B635AF29F2BA046D3A752C26E9D0647B9647D1F4C04AD42000000000000000010000000200000020000000A71D76FAA2D2D5C3224EC3D41DEB293973564A791E55C6782BA76C2BF0495F9A2100000001217DA6C6B3E19F1825CFB2676DAECCE3BF3DE03CF26647C78DF00B371B25CC970000000020000000C4C63F80C74B11263E421EBF8486A4E398D0DBC09FA7D4F62CCDB309F3AEA81F0900000001217DA6C6B3E19F180100000004000000CAFED00D00000000000000000000000000000000FFFFFFFFFFFFFFFF'
    tx = RawTransaction.deserialize(bytes.fromhex(lcs))
    assert tx.__str__() == """{
  "sender": "c3398a599a6f3b9f30b635af29f2ba046d3a752c26e9d0647b9647d1f4c04ad4",
  "sequence_number": 32,
  "payload": {
    "WriteSet": {
      "write_set": [
        [
          {
            "address": "a71d76faa2d2d5c3224ec3d41deb293973564a791e55c6782ba76c2bf0495f9a",
            "path": "01217da6c6b3e19f1825cfb2676daecce3bf3de03cf26647c78df00b371b25cc97"
          },
          "Deletion"
        ],
        [
          {
            "address": "c4c63f80c74b11263e421ebf8486a4e398d0dbc09fa7d4f62ccdb309f3aea81f",
            "path": "01217da6c6b3e19f18"
          },
          {
            "Value": "cafed00d"
          }
        ]
      ]
    }
  },
  "max_gas_amount": 0,
  "gas_unit_price": 0,
  "expiration_time": 18446744073709551615
}"""
    amap = tx.to_json_serializable()
    assert tx.to_json(indent=2) == json.dumps(amap, sort_keys=False, indent=2)
    assert tx.to_json(indent=2) == tx.__str__()




class Cursor:
    def __init__(self, buffer, offset=0):
        self.buffer = buffer
        if isinstance(buffer, list):
            self.buffer = bytes(buffer)
        if isinstance(buffer, bytearray):
            self.buffer = bytes(buffer)
        self.offset = offset
        self.buffer_len = len(self.buffer)

    def read_bytes(self, size):
        end = self.offset + size
        if end > self.buffer_len:
            raise IOError("{} exceed buffer size: {}".format(end, self.buffer_len))
        ret = self.buffer[self.offset:end]
        self.offset = end
        return ret

    def read_to_end(self):
        ret = self.buffer[self.offset:]
        self.offset = self.buffer_len
        return ret

    def peek_bytes(self, size):
        end = self.offset + size
        if end > self.buffer_len:
            raise IOError("{} exceed buffer size: {}".format(end, total))
        return self.buffer[self.offset:end]

    def is_finished(self):
        return self.offset == self.buffer_len

    def position(self):
        return self.offset

    def read_u8(self):
        arr = self.read_bytes(1)
        return int(arr[0])


from canoser.base import Base
from canoser.cursor import Cursor
from canoser.types import type_mapping
from canoser.bool_t import BoolT
from canoser.struct import TypedProperty


class RustOptional(Base):
    _type = None

    def __init__(self, value=None):
        if not self.__class__._type:
            raise TypeError(f'{self.__class__} has no _type defined.')
        self.__dict__["value_type"] = type_mapping(self.__class__._type)
        self.value = value

    def __setattr__(self, name, value):
        if name == "value":
            if value is not None:
                TypedProperty.check_type(self.value_type, value)
            self.__dict__[name] = value
        else:
            raise TypeError(f"{name} not allowed to modify in {self}.")


    @classmethod
    def encode(cls, optional):
        if optional.value is not None:
            ret = BoolT.encode(True)
            ret += optional.value_type.encode(optional.value)
            return ret
        else:
            return BoolT.encode(False)

    @classmethod
    def decode(cls, cursor):
        exist = BoolT.decode(cursor)
        if exist:
            value = cls._type.decode(cursor)
            return cls(value)
        else:
            return cls()

    @classmethod
    def check_value(cls, value):
        if not isinstance(value, cls):
            raise TypeError('value {} is not {} type'.format(value, cls))


    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.value == other.value

    def to_json_serializable(self):
        if self.value is None:
            return None
        return self.value_type.to_json_serializable(self.value)



from canoser import *
from canoser.types import type_mapping
import pdb
import pytest

def test_random():
    for _x in range(10000):
        rand = Uint8.random()
        assert rand >=0 and rand <256

def test_checked_add():
    assert 255 == Uint8.checked_add(254, 1)
    assert None == Uint8.checked_add(254, 2)

def test_str_to_int():
    with pytest.raises(ValueError):
        Uint8.int_unsafe("")
    assert 0 == Uint8.int_unsafe("0")
    assert 0 == Uint8.int_unsafe(b"0")
    assert 1 == Uint8.int_unsafe(b"01")
    assert 0 == Uint8.int_unsafe("0"*100)
    assert 255 == Uint8.int_unsafe("255")
    assert 255 == Uint8.int_unsafe("0255")
    with pytest.raises(TypeError):
        Uint8.int_unsafe("-1")
    with pytest.raises(TypeError):
        Uint8.int_unsafe("256")
    with pytest.raises(TypeError):
        Uint8.check_value(True)


def test_str_to_int_strict():
    assert 0 == Uint8.int_safe(0)
    assert 255 == Uint8.int_safe(255)
    assert -128 == Int8.int_safe(-128)
    assert 127 == Int8.int_safe(127)
    assert 0 == Uint8.int_safe("0")
    assert 255 == Uint8.int_safe("255")
    assert -128 == Int8.int_safe("-128")
    assert 127 == Int8.int_safe("127")
    assert 65535 == Uint16.int_safe("65535")
    with pytest.raises(Exception):
        Uint8.int_safe("")
    with pytest.raises(Exception):
        Uint8.int_safe(b"0")
    with pytest.raises(Exception):
        Uint8.int_safe(b"01")
    with pytest.raises(Exception):
        Uint8.int_safe("0"*100)
    with pytest.raises(Exception):
        Uint8.int_safe("0255")
    with pytest.raises(Exception):
        Uint8.int_safe("-1")
    with pytest.raises(Exception):
        Uint8.int_safe("256")
    with pytest.raises(Exception):
        Int8.int_safe("-129")
    with pytest.raises(Exception):
        Int8.int_safe("128")
    with pytest.raises(Exception):
        Uint8.int_safe(-1)
    with pytest.raises(Exception):
        Uint8.int_safe(256)
    with pytest.raises(Exception):
        Int8.int_safe(-129)
    with pytest.raises(Exception):
        Int8.int_safe(128)

def test_bool_cast_int():
    assert isinstance(True, bool) == True
    assert isinstance(True, int) == True
    assert True == 1
    assert False == (True < 1)
    with pytest.raises(Exception):
        Int8.int_safe(True)
    with pytest.raises(Exception):
        Int8.int_safe(False)
    with pytest.raises(Exception):
        Int8.int_safe(None)

def test_int():
    assert Int8.encode(16) == Uint8.encode(16)


def test_uint8():
    assert Uint8.encode(16) == b"\x10"
    assert Uint8.decode_bytes(b"\x10") == 16
    assert Uint8.max_value == 255
    assert Uint8.min_value == 0

def test_uint8_illegal():
    with pytest.raises(Exception):
        Uint8.encode(-1)
    with pytest.raises(Exception):
        Uint8.encode(0.1)
    with pytest.raises(Exception):
        Uint8.encode([0])
    with pytest.raises(Exception):
        Uint8.encode(b'0')

def test_int8():
    assert Int8.encode(16) == b"\x10"
    assert Int8.decode_bytes(b"\x10") == 16
    assert Int8.max_value == 127
    assert Int8.min_value == -128
    assert Int8.encode(-1) ==  b"\xFF"
    assert Int8.decode_bytes(b"\xFF") == -1
    assert Int8.decode_bytes(b"\x80") == -128


def test_uint16():
    assert Uint16.encode(16) == b"\x10\x00"
    assert Uint16.encode(257) == b"\x01\x01"
    assert Uint16.decode_bytes(b"\x01\x01") == 257
    assert Uint16.max_value == 65535
    assert Uint16.min_value == 0

def test_int16():
    assert Int16.encode(16) == b"\x10\x00"
    assert Int16.decode_bytes(b"\x10\x00") == 16
    assert Int16.max_value == 32767
    assert Int16.min_value == -32768
    assert Int16.encode(-1) ==  b"\xFF\xFF"
    assert Int16.decode_bytes(b"\xFF\xFF") == -1
    assert Int16.decode_bytes(b"\x00\x80") == -32768

def test_uint32():
    assert Uint32.encode(16) == b"\x10\x00\x00\x00"
    assert Uint32.encode(0x12345678) == b"\x78\x56\x34\x12"
    assert Uint32.decode_bytes(b"\x78\x56\x34\x12") == 0x12345678


def test_uint64():
    assert Uint64.encode(16) == b"\x10\x00\x00\x00\x00\x00\x00\x00"
    assert Uint64.encode(0x1234567811223344) == b"\x44\x33\x22\x11\x78\x56\x34\x12"
    assert Uint64.decode_bytes(b"\x44\x33\x22\x11\x78\x56\x34\x12" ) == 0x1234567811223344

def test_uint128():
    assert Uint128.encode(Uint128.max_value) == b"\xff" * 16
    assert Uint128.decode_bytes(b"\xff" * 16) == Uint128.max_value
    v1 = b"\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00"
    assert Uint128.encode(Uint64.max_value+1) == v1
    assert Uint128.decode_bytes(v1) == Uint64.max_value+1
    v2 = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80'
    assert Uint128.encode(Int128.max_value+1) == v2
    assert Uint128.decode_bytes(v2) == Int128.max_value+1

def test_int128():
    v0 = b"\xff" * 15 + b"\x7f"
    assert Int128.encode(Int128.max_value) == v0
    assert Int128.decode_bytes(v0) == Int128.max_value
    v1 = b"\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00"
    assert Int128.encode(Uint64.max_value+1) == v1
    assert Int128.decode_bytes(v1) == Uint64.max_value+1
    v2 = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80'
    assert Int128.encode(Int128.min_value) == v2
    assert Int128.decode_bytes(v2) == Int128.min_value
    assert Int128.encode(-1) == b"\xff" * 16
    assert Int128.decode_bytes(b"\xff" * 16) == -1



def test_bool():
    assert BoolT.encode(True) == b"\1"
    assert BoolT.encode(False) == b"\0"
    assert BoolT.decode_bytes(b"\1") == True
    assert BoolT.decode_bytes(b"\0") == False
    with pytest.raises(TypeError):
    	BoolT.decode_bytes("\x02")


def test_array():
    arrt = ArrayT(Uint8, 2)
    assert arrt.encode([1, 2]) == b'\x02\x00\x00\x00\x01\x02'
    arr = arrt.decode(Cursor(b'\x02\x00\x00\x00\x01\x02'))
    assert arr == [1, 2]
    with pytest.raises(TypeError):
        arrt.decode(Cursor(b'\x01\x00\x00\x00\x01\x02'))
    with pytest.raises(TypeError):
        arrt.decode(Cursor(b'\x03\x00\x00\x00\x01\x02'))

def test_deserialize_int_array():
    arrt = ArrayT(BoolT, 2)
    bools = arrt.decode(Cursor([2,0,0,0,1,0]))
    assert bools == [True, False]

def test_tuple():
    tuplet = TupleT(StrT, Uint8, BoolT)
    assert tuplet.encode(("abc", 1, False)) == b'\x03\x00\x00\x00\x61\x62\x63\x01\x00'
    ret = tuplet.decode(Cursor(b'\x03\x00\x00\x00\x61\x62\x63\x01\x00'))
    assert ret == ("abc", 1, False)

def test_type_mapping():
    atype = {Uint16: Uint64}
    mtype = type_mapping(atype)
    print(mtype)



import pytest
import pdb
from canoser import *

class OptionUInt(RustOptional):
    _type = Uint8

def test_optional():
    null = OptionUInt(None)
    obj = OptionUInt(8)
    assert OptionUInt.encode(null) == b'\x00'
    assert OptionUInt.encode(obj) == b'\x01\x08'
    assert OptionUInt.decode(Cursor(b'\x01\x08')).value == 8
    assert OptionUInt.decode(Cursor(b'\x00')).value == None
    assert obj.serialize() == b'\x01\x08'
    assert obj == OptionUInt.deserialize(obj.serialize())
    with pytest.raises(TypeError):
        obj.value = -1
    with pytest.raises(TypeError):
        obj.value = "abc"
    obj.value = None
    assert obj.value == None
    obj.value = 123
    assert obj.value == 123


class OptionInt(RustOptional):
    _type = Int8

class OStruct(Struct):
    _fields = [('opt', OptionInt)]

def test_optional_struct():
    x = OStruct(opt = OptionInt(-1))
    assert x.to_json_serializable() == {'opt': -1}
    assert OStruct.opt.expected_type == OptionInt
    sx = x.serialize()
    assert sx == b'\x01\xff'
    x2 = OStruct.deserialize(sx)
    assert x.opt.value == x2.opt.value
    with pytest.raises(TypeError):
        x.opt = -1

def test_optional_struct_null():
    x = OStruct(opt = OptionInt())
    assert x.to_json_serializable() == {'opt': None}
    assert x.opt.value is None
    sx = x.serialize()
    assert sx == b'\x00'
    x2 = OStruct.deserialize(sx)
    assert x2.opt.value is None



import struct

def int_list_to_hex(ints):
    return bytes(ints).hex()

def int_list_to_bytes(ints):
    return bytes(ints)
    # return struct.pack("<{}B".format(len(ints)), *ints).hex()

def bytes_to_int_list(bytes_str):
    tp = struct.unpack("<{}B".format(len(bytes_str)), bytes_str)
    return list(tp)

def bytes_to_hex(bytes_str):
    return bytes_str.hex()

def hex_to_bytes(hex_str):
    return bytes.fromhex(hex_str)

def hex_to_int_list(hex_str):
    return bytes_to_int_list(bytes.fromhex(hex_str))


import struct
from random import randint
from canoser.base import Base
from struct import pack, unpack
from_bytes = int.from_bytes

class IntType(Base):

    @classmethod
    def to_json_serializable(cls, value):
        return value

    @classmethod
    def encode(cls, value):
        return pack(cls.pack_str, value)

    @classmethod
    def encode_slow(cls, value):
        return value.to_bytes(cls.byte_lens, byteorder="little", signed=cls.signed)

    @classmethod
    def decode_bytes_slow(cls, bytes):
        return unpack(cls.pack_str, bytes)[0]

    @classmethod
    def decode_bytes(cls, bytes):
        return from_bytes(bytes, byteorder='little', signed=cls.signed)

    @classmethod
    def decode(cls, cursor):
        bytes = cursor.read_bytes(cls.byte_lens)
        return cls.decode_bytes(bytes)

    @classmethod
    def int_unsafe(cls, s):
        ret = int(s)
        cls.check_value(ret)
        return ret

    @classmethod
    def int_safe(cls, s):
        """
        Only allow safe str and valid int to be coerced to destination IntType
        """
        if isinstance(s, bool):
            raise TypeError(f"{s} is not a integer")
        if isinstance(s, int):
            cls.check_value(s)
            return s
        if not isinstance(s, str):
            raise TypeError(f"{s} is not instance of <str>.")
        if len(s) < 1:
            raise TypeError(f"'{s}' is empty.")
        len_min = len(str(cls.min_value))
        len_max = len(str(cls.max_value))
        if len(s) > max(len_min, len_max):
            raise TypeError(f"Length of {s} is larger than max:{max(len_min, len_max)}.")
        ret = int(s)
        cls.check_value(ret)
        return ret


    @classmethod
    def check_value(cls, value):
        if isinstance(value, bool):
            raise TypeError(f"{value} is not a integer")
        if not isinstance(value, int):
            raise TypeError(f"{value} is not instance of <int>.")
        min, max = cls.min_value, cls.max_value
        if value < min or value > max:
            raise TypeError('value {} not in range {}-{}'.format(value, min, max))

    @classmethod
    def checked_add(cls, v1, v2):
        #rust style api
        cls.check_value(v1)
        cls.check_value(v2)
        try:
            ret = v1+v2
            cls.check_value(ret)
            return ret
        except TypeError:
            return None

    @classmethod
    def random(cls):
        return randint(cls.min_value, cls.max_value)

class Int8(IntType):
    pack_str = "<b"
    byte_lens = 1
    max_value = 127
    min_value = -128
    signed = True

class Int16(IntType):
    pack_str = "<h"
    byte_lens = 2
    max_value = 32767
    min_value = -32768
    signed = True

class Int32(IntType):
    pack_str = "<l"
    byte_lens = 4
    max_value = 2147483647
    min_value = -2147483648
    signed = True

class Int64(IntType):
    pack_str = "<q"
    byte_lens = 8
    max_value = 9223372036854775807
    min_value = -9223372036854775808
    signed = True


class Uint8(IntType):
    pack_str = "<B"
    byte_lens = 1
    max_value = 255
    min_value = 0
    signed = False

class Uint16(IntType):
    pack_str = "<H"
    byte_lens = 2
    max_value = 65535
    min_value = 0
    signed = False

class Uint32(IntType):
    pack_str = "<L"
    byte_lens = 4
    max_value = 4294967295
    min_value = 0
    signed = False

class Uint64(IntType):
    pack_str = "<Q"
    byte_lens = 8
    max_value = 18446744073709551615
    min_value = 0
    signed = False


class Int128(IntType):
    byte_lens = 16
    max_value = 170141183460469231731687303715884105727
    min_value = -170141183460469231731687303715884105728
    signed = True

    @classmethod
    def encode(cls, value):
        return value.to_bytes(16, byteorder="little", signed=True)



class Uint128(IntType):
    byte_lens = 16
    max_value = 340282366920938463463374607431768211455
    min_value = 0
    signed = False

    @classmethod
    def encode(cls, value):
        return value.to_bytes(16, byteorder="little", signed=False)




import pytest
import pdb
from copy import copy, deepcopy
from canoser import *

ADDRESS_LENGTH = 32


class TransactionArgument(RustEnum):
    _enums = [
        ('U64', Uint64),
        ('Address', [Uint8, ADDRESS_LENGTH]),
        ('String', str),
        ('ByteArray', [Uint8])
    ]



def test_invalid():
    with pytest.raises(TypeError):
        t_arg = TransactionArgument()
    with pytest.raises(TypeError):
        x = TransactionArgument('UU64', 0)
    with pytest.raises(TypeError):
        class NotDefineEnum(RustEnum):
            pass
        x = NotDefineEnum('X')


def test_enum():
    t_arg = TransactionArgument('U64', 2)
    assert hasattr(t_arg, 'value')
    assert hasattr(t_arg, 'U64')
    assert hasattr(t_arg, 'Address')
    assert hasattr(t_arg, '__str__')
    assert hasattr(t_arg, 'to_proto') == False
    assert t_arg.index == 0
    assert t_arg.value == 2
    assert t_arg.U64 == True
    assert t_arg.enum_name == 'U64'
    assert t_arg.value_type == Uint64
    assert t_arg.serialize() == b"\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00"
    assert t_arg == TransactionArgument.deserialize(t_arg.serialize())
    t_arg.value = 3
    assert t_arg.value == 3
    with pytest.raises(TypeError):
        t_arg.value = 'abc'
    with pytest.raises(TypeError):
        t_arg.index = 2
    arg2 = TransactionArgument('String', 'abc')
    assert arg2.index == 2
    assert arg2.value == 'abc'
    assert arg2.String == True
    assert arg2.U64 == False
    assert arg2.value_type == StrT
    assert arg2.__class__ == t_arg.__class__
    arg2.value = 'Bcd'
    assert arg2.value == 'Bcd'
    with pytest.raises(TypeError):
        t_arg.value = 'abc'
    with pytest.raises(TypeError):
        arg2.value = 0
    with pytest.raises(TypeError):
        arg2.String = 0
    with pytest.raises(TypeError):
        arg2.String = 'abc'

class Enum1(RustEnum):
    _enums = [('opt1', [Uint8]), ('opt2', None)]


def test_enum2():
    e1 = Enum1.new_with_index_value(0, [5])
    e2 = Enum1('opt2', None)
    assert Enum1.new_with_index_value(1, None) == Enum1('opt2')
    assert Enum1.encode(e1) == b'\x00\x00\x00\x00\x01\x00\x00\x00\x05'
    assert Enum1.encode(e2) == b'\x01\x00\x00\x00'
    obj = Enum1.decode(Cursor(b'\x00\x00\x00\x00\x01\x00\x00\x00\x05'))
    assert obj.index == 0
    assert obj.value == [5]
    obj = Enum1.decode(Cursor(b'\x01\x00\x00\x00'))
    assert obj.index == 1
    assert obj.value == None

class Enum2(RustEnum):
    _enums = [('opt1', None), ('opt2', str)]


class EStruct(Struct):
    _fields = [('enum', Enum2)]

def test_enum_struct():
    EStruct.initailize_fields_type()
    assert EStruct.enum.expected_type == Enum2
    x = EStruct(Enum2('opt1'))
    sx = x.serialize()
    assert sx == b'\x00\x00\x00\x00'
    x2 = EStruct.deserialize(sx)
    assert x.enum.index == x2.enum.index
    assert x.enum.value == x2.enum.value


class MyEnum(RustEnum):
    _enums = [('opt1', None), ('opt3', [[str]])]

class EStruct2(Struct):
    _fields = [('enum', MyEnum)]

def test_enum_struct2():
    EStruct2.initailize_fields_type()
    assert EStruct2.enum.expected_type == MyEnum
    x = EStruct2(MyEnum('opt3', [['ab', 'c'], ['d'], []]))
    deepcopy(x)
    sx = x.serialize()
    assert sx == b'\x01\x00\x00\x00\x03\x00\x00\x00' +\
        b'\x02\x00\x00\x00' + b'\x02\x00\x00\x00ab' + b'\x01\x00\x00\x00c' +\
        b'\x01\x00\x00\x00' + b'\x01\x00\x00\x00d' + b'\x00\x00\x00\x00'
    x2 = EStruct2.deserialize(sx)
    assert x.enum.index == x2.enum.index
    assert x.enum.value == x2.enum.value

class MyEnumWrap(MyEnum):
    pass

def test_enum_wrap():
    x1 = MyEnumWrap('opt1')
    x2 = MyEnumWrap('opt3', [['ab', 'c'], ['d'], []])
    assert x2.value == [['ab', 'c'], ['d'], []]
    assert x2.opt3 == True
    assert x2.opt1 == False
    x3 = copy(x2)
    x2.value = [['abc']]
    assert x2.opt3 == True
    assert x2.value == [['abc']]
    assert x3.value == [['ab', 'c'], ['d'], []]
    jstr = x3.to_json()
    print(jstr)
    assert x2.__repr__() == """MyEnumWrap{
  "opt3": [
    [
      "abc"
    ]
  ]
}"""



from hypothesis import given
import hypothesis.strategies as st
from canoser import Struct, Uint8, Uint64, Uint128, Int8, Int64, Int128


def encode_decode_is_same(vtype, value):
	byts = vtype.encode(value)
	value2 = vtype.decode_bytes(byts)
	assert value == value2


@given(st.integers())
def test_prop(x):
	if x >= Uint8.min_value and x <= Uint8.max_value:
		encode_decode_is_same(Uint8, x)
	if x >= Uint64.min_value and x <= Uint64.max_value:
		encode_decode_is_same(Uint64, x)
	if x >= Uint128.min_value and x <= Uint128.max_value:
		encode_decode_is_same(Uint128, x)
	if x >= Int8.min_value and x <= Int8.max_value:
		encode_decode_is_same(Int8, x)
	if x >= Int64.min_value and x <= Int64.max_value:
		encode_decode_is_same(Int64, x)
	if x >= Int128.min_value and x <= Int128.max_value:
		encode_decode_is_same(Int128, x)


class PropStruct(Struct):
    _fields = [('name', str), ('value', Int128), ('flag', bool)]

@given(st.text(), st.integers(), st.booleans())
def test_ser_deser(name, value, flag):
	t = PropStruct(name, value, flag)
	ser = t.serialize()
	t2 = PropStruct.deserialize(ser)
	assert t == t2



from canoser.cursor import Cursor  # noqa: F401
from canoser.struct import Struct  # noqa: F401
from canoser.rust_enum import RustEnum  # noqa: F401
from canoser.rust_optional import RustOptional  # noqa: F401
from canoser.delegate_t import DelegateT  # noqa: F401
from canoser.tuple_t import TupleT  # noqa: F401
from canoser.map_t import MapT  # noqa: F401
from canoser.str_t import StrT  # noqa: F401
from canoser.bytes_t import BytesT  # noqa: F401
from canoser.bool_t import BoolT  # noqa: F401
from canoser.array_t import ArrayT  # noqa: F401
from canoser.int_type import Uint8, Uint16, Uint32, Uint64, Int8, Int16, Int32, Int64  # noqa: F401
from canoser.int_type import Uint128, Int128  # noqa: F401
from canoser.util import bytes_to_int_list, hex_to_int_list  # noqa: F401



from canoser.base import Base

class TupleT(Base):

    def __init__(self, *ttypes):
        self.ttypes = ttypes

    def encode(self, value):
        output = b""
        zipped = zip(self.ttypes, value)
        for k, v in zipped:
            output += k.encode(v)
        return output

    def decode(self, cursor):
        arr = []
        for k in self.ttypes:
            arr.append(k.decode(cursor))
        return tuple(arr)

    def check_value(self, value):
        if len(value) != len(self.ttypes):
            raise TypeError(f"{len(value)} is not equal to {len(self.ttypes)}")
        if not isinstance(value, tuple):
            raise TypeError(f"{value} is not a tuple.")
        zipped = zip(self.ttypes, value)
        for k, v in zipped:
            k.check_value(v)

    def __eq__(self, other):
        if not isinstance(other, TupleT):
            return False
        zipped = zip(self.ttypes, other.ttypes)
        for t1, t2 in zipped:
            if t1 != t2:
                return False
        return True

    def to_json_serializable(cls, obj):
        ret = []
        #https://stackoverflow.com/questions/15721363/preserve-python-tuples-with-json
        #If need to deserialize tuple back later, above link will help.
        zipped = zip(cls.ttypes, obj)
        for k, v in zipped:
            data = k.to_json_serializable(v)
            ret.append(data)
        return ret



from canoser import *
import pytest
import pdb

ADDRESS_LENGTH = 32

input = [
        0xca, 0x82, 0x0b, 0xf9, 0x30, 0x5e, 0xb9, 0x7d, 0x0d, 0x78, 0x4f, 0x71, 0xb3, 0x95, 0x54,
        0x57, 0xfb, 0xf6, 0x91, 0x1f, 0x53, 0x00, 0xce, 0xaa, 0x5d, 0x7e, 0x86, 0x21, 0x52, 0x9e,
        0xae, 0x19,
    ]

expected_output = [
        0xCA, 0x82, 0x0B, 0xF9, 0x30, 0x5E, 0xB9, 0x7D, 0x0D, 0x78, 0x4F,
        0x71, 0xB3, 0x95, 0x54, 0x57, 0xFB, 0xF6, 0x91, 0x1F, 0x53, 0x00, 0xCE, 0xAA, 0x5D, 0x7E,
        0x86, 0x21, 0x52, 0x9E, 0xAE, 0x19,
    ]


class Address1(DelegateT):
    delegate_type = [Uint8, ADDRESS_LENGTH]


class Address2(DelegateT):
    delegate_type = [Uint8, ADDRESS_LENGTH, False]


def test_enocde_len():
    expected_output0 = [32, 0, 0, 0] + expected_output
    actual_output = Address1.encode(input)
    assert bytes(expected_output0) == actual_output

def test_not_enocde_len():
    actual_output = Address2.encode(input)
    assert bytes(input) == actual_output
    assert bytes(expected_output) == actual_output

class AddrStruct(Struct):
    _fields = [('map', {Address2: [str]})]

def test_int_list_as_dict_key():
    amap = {bytes(input): ['test']}
    addrs = AddrStruct(amap)
    ser = addrs.serialize()
    addr2 = AddrStruct.deserialize(ser)
    assert addrs.map == addr2.map


from canoser import *
import pytest


ADDRESS_LENGTH = 32

input = b'\xca\x82\x0b\xf90^\xb9}\rxOq\xb3\x95TW\xfb\xf6\x91\x1fS\x00\xce\xaa]~\x86!R\x9e\xae\x19'

class Address1(DelegateT):
    delegate_type = BytesT(ADDRESS_LENGTH)


class Address2(DelegateT):
    delegate_type = BytesT(ADDRESS_LENGTH, encode_len=False)


def test_enocde_len():
    expected_output = bytes([32, 0, 0, 0]) + input
    actual_output = Address1.encode(input)
    assert expected_output == actual_output

def test_not_enocde_len():
    actual_output = Address2.encode(input)
    assert input == actual_output

class AddrStruct(Struct):
    _fields = [('map', {Address2: [str]})]

def test_address_as_dict_key():
    amap = {input: ['test']}
    addrs = AddrStruct(amap)
    ser = addrs.serialize()
    addr2 = AddrStruct.deserialize(ser)
    assert addrs.map == addr2.map


from canoser import *
import struct
import pdb
import pytest

ADDRESS_LENGTH = 32

class Address(DelegateT):
    delegate_type = [Uint8, ADDRESS_LENGTH]


class AddrStruct(Struct):
    _fields = [('addr', Address)]

def test_array():
    with pytest.raises(TypeError):
        x = AddrStruct([1,2,3])
    arr = []
    for _ in range(ADDRESS_LENGTH):
        arr.append(5)
    x = AddrStruct(arr)
    sx = x.serialize()
    x2 = AddrStruct.deserialize(sx)
    assert x.addr == x2.addr
    print(x2)
    assert x2.__str__() == """{
  "addr": "0505050505050505050505050505050505050505050505050505050505050505"
}"""
    assert AddrStruct.addr.expected_type.dtype().atype == Uint8
    assert AddrStruct.addr.expected_type.dtype().fixed_len == ADDRESS_LENGTH


class Bools(DelegateT):
    delegate_type = [bool]

def test_delegate():
    x = [True, False, True]
    bs = Bools.encode(x)
    assert bs == b'\x03\x00\x00\x00\x01\x00\x01'
    x2 = Bools.deserialize(bs)
    assert x == x2



from canoser.base import Base
from canoser.int_type import Uint32

class MapT(Base):

    def __init__(self, ktype, vtype):
        self.ktype = ktype
        self.vtype = vtype

    def encode(self, kvs):
        output = b""
        output += Uint32.encode(len(kvs))
        odict = {}
        for k, v in kvs.items():
            odict[self.ktype.encode(k)] = self.vtype.encode(v)
        for name in sorted(odict.keys()):
            output += name
            output += odict[name]
        return output

    def decode(self, cursor):
        kvs = {}
        size = Uint32.decode(cursor)
        for _ in range(size):
            k = self.ktype.decode(cursor)
            v = self.vtype.decode(cursor)
            if isinstance(k, list) and isinstance(k[0], int):
                #python doesn't support list as key in dict, so we change list to bytes
                kvs[bytes(k)] = v
            else:
                kvs[k] = v
        #TODO: check the key order of kvs, because lcs has order when serialize map.
        return kvs

    def check_value(self, kvs):
        if not isinstance(kvs, dict):
            raise TypeError(f"{kvs} is not a dict.")
        for k, v in kvs.items():
            if isinstance(self.ktype, list) or \
                (hasattr(self.ktype, 'delegate_type') and isinstance(self.ktype.delegate_type, list)):
                from canoser.types import BytesT
                BytesT().check_value(k)
            else:
                self.ktype.check_value(k)
            self.vtype.check_value(v)

    def __eq__(self, other):
        if not isinstance(other, MapT):
            return False
        return self.ktype == other.ktype and self.vtype == other.vtype

    def to_json_serializable(cls, obj):
        amap = {}
        for k,v in obj.items():
            kk = cls.ktype.to_json_serializable(k)
            vv = cls.vtype.to_json_serializable(v)
            amap[kk] = vv
        return amap





from canoser.types import type_mapping
from canoser.base import Base


class DelegateT(Base):
    delegate_type = 'delegate'

    @classmethod
    def dtype(cls):
        return type_mapping(cls.delegate_type)

    @classmethod
    def encode(cls, value):
        return cls.dtype().encode(value)

    @classmethod
    def decode(cls, cursor):
        return cls.dtype().decode(cursor)

    @classmethod
    def check_value(cls, value):
        cls.dtype().check_value(value)

    @classmethod
    def to_json_serializable(cls, value):
        #TODO: bad smell
        #if hasattr(cls, "to_json_serializable"):
        if 'to_json_serializable' in cls.__dict__.keys():
            return cls.to_json_serializable(value)
        return cls.dtype().to_json_serializable(value)



from canoser.int_type import Uint8
from canoser.tuple_t import TupleT
from canoser.map_t import MapT
from canoser.str_t import StrT
from canoser.bytes_t import BytesT
from canoser.bool_t import BoolT
from canoser.array_t import ArrayT


def my_import(name):
    components = name.split('.')
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod

def type_mapping(field_type):
    """
    Mapping python types to canoser types
    """
    if field_type == str:
        return StrT
    elif field_type == bytes:
        return BytesT()
    elif field_type == bool:
        return BoolT
    elif type(field_type) == list:
        if len(field_type) == 0:
            return ArrayT(Uint8)
        elif len(field_type) == 1:
            item = field_type[0]
            return ArrayT(type_mapping(item))
        elif len(field_type) == 2:
            item = field_type[0]
            size = field_type[1]
            return ArrayT(type_mapping(item), size)
        elif len(field_type) == 3:
            item = field_type[0]
            size = field_type[1]
            encode_len = field_type[2]
            return ArrayT(type_mapping(item), size, encode_len)
        else:
            raise TypeError("Array has one item type, no more.")
        raise AssertionError("unreacheable")
    elif type(field_type) == dict:
        if len(field_type) == 0:
            ktype = BytesT()
            vtype = [Uint8]
        elif len(field_type) == 1:
            ktype = next(iter(field_type.keys()))
            vtype = next(iter(field_type.values()))
        else:
            raise TypeError("Map type has one item mapping key type to value type.")
        return MapT(type_mapping(ktype), type_mapping(vtype))
    elif type(field_type) == tuple:
        arr = []
        for item in field_type:
            arr.append(type_mapping(item))
        return TupleT(*arr)
    elif type(field_type) == str:
        return my_import(field_type)
    else:
        return field_type


from canoser.base import Base

class BoolT(Base):
    @classmethod
    def encode(self, value):
        if value:
            return b'\1'
        else:
            return b'\0'

    @classmethod
    def decode_bytes(self, value):
        if value == b'\0':
            return False
        elif value == b'\1':
            return True
        else:
            raise TypeError("bool should be 0 or 1.")

    @classmethod
    def decode(self, cursor):
        value = cursor.read_bytes(1)
        return self.decode_bytes(value)

    @classmethod
    def check_value(self, value):
        if not isinstance(value, bool):
            raise TypeError('value {} is not bool'.format(value))

    @classmethod
    def to_json_serializable(cls, value):
        return value



from canoser.base import Base
from canoser.cursor import Cursor
from canoser.types import type_mapping
from canoser.int_type import Uint32
from canoser.struct import TypedProperty
import json

#TODO: how to support discontinuous index in enum

class RustEnum(Base):
    _enums = []

    @classmethod
    def get_index(cls, name):
        for index, (ename, _) in enumerate(cls._enums):
            if ename == name:
                return index
        raise TypeError(f"name:{name} not in enum {cls}")

    @classmethod
    def new_with_index_value(cls, index, value):
        if not cls._enums:
            raise TypeError(f'{cls} has no _enums defined.')
        if index < 0 or index >= len(cls._enums):
            raise TypeError(f"index{index} out of bound:0-{len(cls._enums)-1}")
        _name, datatype = cls._enums[index]
        ret = cls.__new__(cls)
        ret._init_with_index_value(index, value, datatype)
        return ret

    def _init_with_index_value(self, index, value, datatype):
        self._index = index
        self.value_type = type_mapping(datatype)
        self.value = value

    def __init__(self, name, value=None):
        if not self.__class__._enums:
            raise TypeError(f'{self.__class__} has no _enums defined.')
        index = self.__class__.get_index(name)
        _name, datatype = self._enums[index]
        if name != _name:
            raise AssertionError(f"{name} != {_name}")
        self._init_with_index_value(index, value, datatype)

    #__getattr__ only gets called for attributes that don't actually exist.
    #If you set an attribute directly, referencing that attribute will retrieve it without calling __getattr__.
    #If you need to catch every attribute regardless whether it exists or not, use __getattribute__ instead.
    def __getattr__(self, name):
        if name == '_index':
            return None
        try:
            return self._index == self.__class__.get_index(name)
        except TypeError:
            return super().__getattr__(self, name)

    def __setattr__(self, name, value):
        if name == "value":
            TypedProperty.check_type(self.value_type, value)
            self.__dict__[name] = value
        elif name == "_index" or name == "value_type":
            self.__dict__[name] = value
        else:
            raise TypeError(f"{name} not allowed to modify in {self}.")

    @property
    def index(self):
        return self._index

    @property
    def enum_name(self):
        name, _ = self.__class__._enums[self._index]
        return name

    @classmethod
    def encode(cls, enum):
        ret = Uint32.encode(enum.index)
        if enum.value_type is not None:
            ret += enum.value_type.encode(enum.value)
        return ret

    @classmethod
    def decode(cls, cursor):
        index = Uint32.decode(cursor)
        _name, datatype = cls._enums[index]
        if datatype is not None:
            value = type_mapping(datatype).decode(cursor)
            return cls.new_with_index_value(index, value)
        else:
            return cls.new_with_index_value(index, None)

    @classmethod
    def check_value(cls, value):
        if not isinstance(value, cls):
            raise TypeError('value {} is not {} type'.format(value, cls))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.index == other.index and self.value == other.value

    def to_json_serializable(self):
        if self.value_type == None:
            return self.enum_name
        jj = self.value_type.to_json_serializable(self.value)
        return {self.enum_name : jj}

    def __str__(self):
        return self.to_json(indent=2)

    def __repr__(self):
        return self.__class__.__qualname__ + self.to_json(indent=2)

    def to_json(self, sort_keys=False, indent=4):
        amap = self.to_json_serializable()
        return json.dumps(amap, sort_keys=sort_keys, indent=indent)




from canoser.base import Base
from canoser.cursor import Cursor
from canoser.types import type_mapping
import json


class TypedProperty:
    def __init__(self, name, expected_type):
        self.name = name
        self.expected_type = expected_type

    def __set__(self, instance, value):
        TypedProperty.check_type(self.expected_type, value)
        instance.__dict__[self.name] = value

    @staticmethod
    def check_type(datatype, value):
        if datatype is None:
            if value is not None:
                raise TypeError(f'{datatype} mismatch {value}')
            else:
                return
        check = getattr(datatype, "check_value", None)
        if callable(check):
            check(value)
        else:
            raise TypeError('{} has no check_value method'.format(datatype))



class Struct(Base):
    _fields = []
    _initialized = False

    @classmethod
    def initailize_fields_type(cls):
        if not cls._initialized:
            cls._initialized = True
            for name, atype in cls._fields:
                setattr(cls, name, TypedProperty(name, type_mapping(atype)))

    def __init__(self, *args, **kwargs):
        self.__class__.initailize_fields_type()

        if len(args) > len(self._fields):
            raise TypeError('Expected {} arguments'.format(len(self._fields)))

        # Set all of the positional arguments
        for (name, _type), value in zip(self._fields, args):
            typed = getattr(self, name)
            typed.__set__(self, value)

        # Set the remaining keyword arguments
        for name, _type in self._fields[len(args):]:
            if name in kwargs:
                typed = getattr(self, name)
                typed.__set__(self, kwargs.pop(name))

        # Check for any remaining unknown arguments
        if kwargs:
            raise TypeError('Invalid argument(s): {}'.format(','.join(kwargs)))


    @classmethod
    def encode(cls, obj):
        output = b''
        for name, atype in obj._fields:
            value = getattr(obj, name)
            output += type_mapping(atype).encode(value)
        return output

    @classmethod
    def decode(cls, cursor):
        ret = cls.__new__(cls)
        ret.__init__()
        for name, atype in ret._fields:
            prop = getattr(ret, name)
            mtype = type_mapping(atype)
            assert mtype == prop.expected_type
            value = mtype.decode(cursor)
            prop.__set__(ret, value)
        return ret

    @classmethod
    def check_value(cls, value):
        if not isinstance(value, cls):
            raise TypeError('value {} is not {} type'.format(value, cls))

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        for name, atype in self._fields:
            v1 = getattr(self, name)
            v2 = getattr(other, name)
            if v1 != v2:
                return False
        return True

    def to_json_serializable(self):
        amap = {}
        for name, atype in self._fields:
            value = getattr(self, name)
            if isinstance(value, TypedProperty):
                amap[name] = None
            else:
                atype = type_mapping(atype)
                amap[name] = atype.to_json_serializable(value)
        return amap

    def __str__(self):
        return self.to_json(indent=2)

    def __repr__(self):
        return self.__class__.__qualname__ + self.to_json(indent=2)

    def to_json(self, sort_keys=False, indent=4):
        amap = self.to_json_serializable()
        return json.dumps(amap, sort_keys=sort_keys, indent=indent)




from canoser import *
import pytest
import pdb

class Circular(Struct):
    _fields = [
        ('intv', Int8),
        ('next', ['test_circular.Circular'])
    ]

def test_circular():
    t1 = Circular(1,[])
    t2 = Circular(2,[])
    t12 = Circular(12,[t1, t2])
    bstr = t12.serialize()
    assert bstr == bytes([12]) + Uint32.encode(2) + t1.serialize() + t2.serialize()
    tt = Circular.deserialize(bstr)
    assert tt == t12
    assert Circular.deserialize(t1.serialize()) == t1



from canoser.base import Base
from canoser.int_type import Uint32, Uint8
import struct


class ArrayT(Base):

    def __init__(self, atype, fixed_len=None, encode_len=True):
        self.atype = atype
        if fixed_len is not None and fixed_len <= 0:
            raise TypeError("arr len must > 0".format(fixed_len))
        if fixed_len is None and not encode_len:
            raise TypeError("variable length sequences must encode len.")
        self.fixed_len = fixed_len
        self.encode_len = encode_len

    def encode(self, arr):
        if self.fixed_len is not None and len(arr) != self.fixed_len:
             raise TypeError(f"{len(arr)} is not equal to predefined value: {self.fixed_len}")
        output = b""
        if self.encode_len:
            output += Uint32.encode(len(arr))
        for item in arr:
            output += self.atype.encode(item)
        return output

    def decode(self, cursor):
        arr = []
        if self.encode_len:
            size = Uint32.decode(cursor)
            if self.fixed_len is not None and size != self.fixed_len:
                 raise TypeError(f"{size} is not equal to predefined value: {self.fixed_len}")
        else:
            size = self.fixed_len
        for _ in range(size):
            arr.append(self.atype.decode(cursor))
        return arr

    def check_value(self, arr):
        if self.fixed_len is not None and len(arr) != self.fixed_len:
            raise TypeError("arr len not match: {}-{}".format(len(arr), self.fixed_len))
        if not isinstance(arr, list):
            raise TypeError(f"{arr} is not a list.")
        for item in arr:
            self.atype.check_value(item)

    def __eq__(self, other):
        if not isinstance(other, ArrayT):
            return False
        return self.atype == other.atype and self.fixed_len == other.fixed_len

    def to_json_serializable(cls, obj):
        if cls.atype == Uint8:
            return struct.pack("<{}B".format(len(obj)), *obj).hex()
        ret = []
        for _, item in enumerate(obj):
            data = cls.atype.to_json_serializable(item)
            ret.append(data)
        return ret




from canoser.int_type import Uint32
from canoser.base import Base

class StrT(Base):
    @classmethod
    def encode(self, value):
        output = b''
        utf8 = value.encode('utf-8')
        output += Uint32.encode(len(utf8))
        output += utf8
        return output

    @classmethod
    def decode(self, cursor):
        strlen = Uint32.decode(cursor)
        return str(cursor.read_bytes(strlen), encoding='utf-8')

    @classmethod
    def check_value(cls, value):
        if not isinstance(value, str):
            raise TypeError('value {} is not string'.format(value))

    @classmethod
    def to_json_serializable(cls, obj):
        return obj





import setuptools
from canoser.version import version

with open("README.md", "r") as fh:
    content = fh.read()
    arr = content.split("\n")
    long_description = "\n".join(arr[4:])

setuptools.setup(
    name="canoser",
    version=version,
    author="yuan xinyu",
    author_email="yuan_xin_yu@hotmail.com",
    description="A python implementation of the LCS(Libra Canonical Serialization) for the Libra network.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yuan-xy/canoser-python.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from canoser import *
import pytest
import pdb

def test_read():
    data = [6,2,3,4,5]
    cursor = Cursor(data)
    assert cursor.read_u8() == 6
    assert cursor.offset == 1
    assert cursor.position() == cursor.offset
    assert cursor.peek_bytes(3) == b'\x02\x03\x04'
    assert cursor.offset == 1
    assert cursor.read_bytes(2) == b'\x02\x03'
    assert cursor.offset == 3
    assert cursor.is_finished() == False
    assert cursor.read_to_end() == b'\x04\x05'
    assert cursor.is_finished() == True

def test_bytearray():
    array = bytearray()
    array.append(5)
    array.append(2)
    array.extend([3, 4])
    cursor = Cursor(array)
    assert cursor.read_u8() == 5
    assert cursor.offset == 1
    assert cursor.position() == cursor.offset
    assert cursor.peek_bytes(3) == b'\x02\x03\x04'
    assert cursor.offset == 1
    assert cursor.read_bytes(2) == b'\x02\x03'
    assert cursor.offset == 3
    assert cursor.is_finished() == False
    assert cursor.read_to_end() == b'\x04'
    assert cursor.is_finished() == True


import pytest
import pdb
from canoser import *


class Stock(Struct):
    _fields = [('name', str), ('shares', Uint8)]


def test_struct_init():
    s1 = Stock('ACME', 50)
    assert s1.name == "ACME"
    assert s1.shares == 50
    s2 = Stock('ACME', shares=50)
    s3 = Stock(name='ACME', shares=50)
    s33 = Stock(shares=50, name='ACME')
    assert s3 == s33
    s6 = Stock(shares=50)
    with pytest.raises(TypeError):
        s4 = Stock('ACME', 500)
    with pytest.raises(TypeError):
        s5 = Stock('ACME', shares=50, aa=1)
    with pytest.raises(TypeError):
        s6 = Stock(123)


def test_struct_serialize():
    s1 = Stock('ACME', 50)
    bstr = s1.serialize()
    s2 = Stock.deserialize(bstr)
    assert s2.name == "ACME"
    assert s2.shares == 50

class BoolS(Struct):
    _fields = [('boolean', bool)]

def test_bool():
    x = BoolS(True)
    sx = x.serialize()
    x2 = BoolS.deserialize(sx)
    assert x.boolean == x2.boolean
    assert x.to_json() == """{
    "boolean": true
}"""

class ArrayS(Struct):
    _fields = [('array', [bool])]

def test_array():
    x = ArrayS([True, False, True])
    sx = x.serialize()
    assert b"\3\0\0\0\1\0\1" == sx
    x2 = ArrayS.deserialize(sx)
    assert x.array == x2.array

def test_array_error():
    with pytest.raises(TypeError):
        ArrayS.deserialize(b"\3\0\0\0\1\0\2")
    x = ArrayS([])
    with pytest.raises(TypeError):
        x.array = ["abc"]
    # with pytest.raises(TypeError):
    #     x.array.append("abc")


class ArrayUint8(Struct):
    _fields = [('array', [Uint8])]

def test_array():
    with pytest.raises(Exception):
        x = ArrayUint8([0.1])
    with pytest.raises(TypeError):
        x = ArrayUint8([-1])
    with pytest.raises(TypeError):
        x = ArrayUint8(b'')
    x = ArrayUint8([])
    with pytest.raises(TypeError):
        x.array = b''


class MapS(Struct):
    _fields = [('kvs', {str : Uint64})]

def test_map():
    x = MapS(kvs = {"count1":123456789, "count2":987654321})
    sx = x.serialize()
    x2 = MapS.deserialize(sx)
    assert x.kvs == x2.kvs

class MapS2(Struct):
    _fields = [('counters', {Uint16 : Uint64})]


def test_map_empty():
    x = MapS2({})
    sx = x.serialize()
    x2 = MapS2.deserialize(sx)
    assert x.counters == x2.counters

class ChineseMap(Struct):
    _fields = [('kvs', {str : str})]

def test_map2():
    x = ChineseMap(kvs = {"中文":"测试"})
    sx = x.serialize()
    x2 = ChineseMap.deserialize(sx)
    assert x.kvs == x2.kvs
    assert x2.kvs["中文"] == "测试"
    assert x2.to_json() == """{
    "kvs": {
        "\\u4e2d\\u6587": "\\u6d4b\\u8bd5"
    }
}"""



class ByteS(Struct):
    _fields = [('kvs', {bytes : Uint64})]

def test_bytes():
    x = ByteS(kvs = {b"count1":123456789, b"count2":987654321})
    sx = x.serialize()
    x2 = ByteS.deserialize(sx)
    assert x.kvs == x2.kvs
    with pytest.raises(TypeError):
        x2.kvs = {'a' : 'b'}
    print(x2.to_json())
    assert x2.to_json() == """{
    "kvs": {
        "636f756e7431": 123456789,
        "636f756e7432": 987654321
    }
}"""


class TupleS(Struct):
    _fields = [('tp', (str, Uint8, bool, Int16))]

def test_tuple_struct():
    x = TupleS(tp = ("abc", 1, False, 2))
    assert TupleS.tp.expected_type == TupleT(StrT, Uint8, BoolT, Int16)
    sx = x.serialize()
    assert sx == b'\x03\x00\x00\x00\x61\x62\x63\x01\x00\x02\x00'
    x2 = TupleS.deserialize(sx)
    assert x.tp == x2.tp
    with pytest.raises(TypeError):
        x.tp = ()



class Uint128S(Struct):
    _fields = [('u128', Uint128)]

def test_uint128_struct():
    x = Uint128S(u128 = Uint128.max_value)
    assert Uint128S.u128.expected_type == Uint128
    sx = x.serialize()
    assert sx == b'\xff' * 16
    x2 = Uint128S.deserialize(sx)
    assert x.u128 == x2.u128

def test_print_null_field():
    x = Uint128S()
    print(x)
    print(x.__repr__())


import struct
from canoser.int_type import Uint32
from canoser.base import Base

class BytesT(Base):

    def __init__(self, fixed_len=None, encode_len=True):
        if fixed_len is not None and fixed_len <= 0:
            raise TypeError("byte len must > 0".format(fixed_len))
        if fixed_len is None and not encode_len:
            raise TypeError("variable length sequences must encode len.")
        self.fixed_len = fixed_len
        self.encode_len = encode_len


    def encode(self, value):
        output = b""
        if self.encode_len:
            output += Uint32.encode(len(value))
        output += value
        return output


    def decode(self, cursor):
        if self.encode_len:
            size = Uint32.decode(cursor)
            if self.fixed_len is not None and size != self.fixed_len:
                 raise TypeError(f"{size} is not equal to predefined value: {self.fixed_len}")
        else:
            size = self.fixed_len
        return cursor.read_bytes(size)

    def check_value(self, value):
        if not isinstance(value, bytes):
            raise TypeError('value {} is not bytes'.format(value))
        if self.fixed_len is not None and len(value) != self.fixed_len:
            raise TypeError("len not match: {}-{}".format(len(value), self.fixed_len))


    def __eq__(self, other):
        if not isinstance(other, BytesT):
            return False
        return self.fixed_len == other.fixed_len and self.encode_len == other.encode_len


    def to_json_serializable(cls, obj):
        return obj.hex()





from canoser.cursor import Cursor
from io import StringIO
import json

class Base:
    """
    All types should implment following four methods:

    def encode(cls_or_obj, value)

    def decode(cls_or_obj, cursor)

    def check_value(cls_or_obj, value)

    def to_json_serializable(obj)
    def to_json_serializable(acls, value)
    """


    def serialize(self):
        return self.__class__.encode(self)

    @classmethod
    def deserialize(cls, buffer, check=True):
        cursor = Cursor(buffer)
        ret = cls.decode(cursor)
        if not cursor.is_finished() and check:
            raise IOError("bytes not all consumed:{}, {}".format(
                len(buffer), cursor.offset))
        return ret

